/*
 *  Created by Martin on 21/11/2017.
 *
 *  This file collects declaration that we want to expose to test files.
 *  These declarations are expected to be duplicated elsewhere,
 *  together with their implementation.
 *
 *  Distributed under the Boost Software License, Version 1.0. (See accompanying
 *  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 */
#ifndef TWOBLUECUBES_CATCH_USER_INTERFACES_H_INCLUDED
#define TWOBLUECUBES_CATCH_USER_INTERFACES_H_INCLUDED

namespace Catch {
    unsigned int rngSeed();
}

#endif // TWOBLUECUBES_CATCH_USER_INTERFACES_H_INCLUDED
