// This file is only here to verify (to the extent possible) the self sufficiency of the header
#include "internal/catch_suppress_warnings.h"
#include "internal/catch_xmlwriter.h"
#include "internal/catch_reenable_warnings.h"
