//
//  TestObj.m
//  OCTest
//
//  Created by Phil on 13/11/2010.
//  Copyright 2010 Two Blue Cubes Ltd. All rights reserved.
//
//  Distributed under the Boost Software License, Version 1.0. (See accompanying
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#import "TestObj.h"


@implementation TestObj

@synthesize int_val;

-(BOOL) isTrue {
    return YES;
}
-(BOOL) isFalse {
    return NO;
}

@end
