/*
 *  Created by Phil on 10/2/2014.
 *  Copyright 2014 Two Blue Cubes Ltd. All rights reserved.
 *
 *  Distributed under the Boost Software License, Version 1.0. (See accompanying
 *  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 */
#ifndef TWOBLUECUBES_CATCH_TEXT_H_INCLUDED
#define TWOBLUECUBES_CATCH_TEXT_H_INCLUDED

#include "catch_clara.h"

namespace Catch {
    using namespace clara::TextFlow;
}

#endif // TWOBLUECUBES_CATCH_TEXT_H_INCLUDED
