// This file #includes all the .cpp files into a single .mm
// - so they get compiled as ObjectiveC++

#include "../../../include/internal/catch_tostring.cpp"
#include "../../../include/internal/catch_approx.cpp"
#include "../../../include/internal/catch_assertionhandler.cpp"
#include "../../../include/internal/catch_assertionresult.cpp"
#include "../../../include/internal/catch_benchmark.cpp"
#include "../../../include/internal/catch_capture_matchers.cpp"
#include "../../../include/internal/catch_commandline.cpp"
#include "../../../include/internal/catch_common.cpp"
#include "../../../include/internal/catch_config.cpp"
#include "../../../include/internal/catch_console_colour.cpp"
#include "../../../include/internal/catch_context.cpp"
#include "../../../include/internal/catch_debug_console.cpp"
#include "../../../include/internal/catch_debugger.cpp"
#include "../../../include/internal/catch_decomposer.cpp"
#include "../../../include/internal/catch_errno_guard.cpp"
#include "../../../include/internal/catch_exception_translator_registry.cpp"
#include "../../../include/internal/catch_fatal_condition.cpp"
#include "../../../include/internal/catch_interfaces_capture.cpp"
#include "../../../include/internal/catch_interfaces_config.cpp"
#include "../../../include/internal/catch_interfaces_exception.cpp"
#include "../../../include/internal/catch_interfaces_registry_hub.cpp"
#include "../../../include/internal/catch_interfaces_reporter.cpp"
#include "../../../include/internal/catch_interfaces_runner.cpp"
#include "../../../include/internal/catch_interfaces_testcase.cpp"
#include "../../../include/internal/catch_leak_detector.cpp"
#include "../../../include/internal/catch_list.cpp"
#include "../../../include/internal/catch_matchers.cpp"
#include "../../../include/internal/catch_matchers_string.cpp"
#include "../../../include/internal/catch_message.cpp"
#include "../../../include/internal/catch_output_redirect.cpp"
#include "../../../include/internal/catch_random_number_generator.cpp"
#include "../../../include/internal/catch_registry_hub.cpp"
#include "../../../include/internal/catch_reporter_registry.cpp"
#include "../../../include/internal/catch_result_type.cpp"
#include "../../../include/internal/catch_run_context.cpp"
#include "../../../include/internal/catch_section.cpp"
#include "../../../include/internal/catch_section_info.cpp"
#include "../../../include/internal/catch_session.cpp"
#include "../../../include/internal/catch_startup_exception_registry.cpp"
#include "../../../include/internal/catch_stream.cpp"
#include "../../../include/internal/catch_string_manip.cpp"
#include "../../../include/internal/catch_stringref.cpp"
#include "../../../include/internal/catch_tag_alias.cpp"
#include "../../../include/internal/catch_tag_alias_autoregistrar.cpp"
#include "../../../include/internal/catch_tag_alias_registry.cpp"
#include "../../../include/internal/catch_test_case_info.cpp"
#include "../../../include/internal/catch_test_case_registry_impl.cpp"
#include "../../../include/internal/catch_test_case_tracker.cpp"
#include "../../../include/internal/catch_test_registry.cpp"
#include "../../../include/internal/catch_test_spec.cpp"
#include "../../../include/internal/catch_test_spec_parser.cpp"
#include "../../../include/internal/catch_timer.cpp"
#include "../../../include/internal/catch_totals.cpp"
#include "../../../include/internal/catch_uncaught_exceptions.cpp"
#include "../../../include/internal/catch_version.cpp"
#include "../../../include/internal/catch_wildcard_pattern.cpp"
#include "../../../include/internal/catch_xmlwriter.cpp"


// Reporters
#include "../../../include/reporters/catch_reporter_bases.cpp"
#include "../../../include/reporters/catch_reporter_compact.cpp"
#include "../../../include/reporters/catch_reporter_console.cpp"
#include "../../../include/reporters/catch_reporter_junit.cpp"
#include "../../../include/reporters/catch_reporter_listening.cpp"
#include "../../../include/reporters/catch_reporter_xml.cpp"
